/* 
 * \file   kh3luaparser.cpp : lua parser for Khepera 3
 * \brief 
 *         This application executes a lua script given in parameter from the V-REP simulation software 
 *         on the real khepera 3 with gripper.
 *        
 * \author   Julien Tharin (K-Team SA)                               
 *
 * \note     Copyright (C) 2010 K-TEAM SA
 * \bug      none discovered. 
 *
 * \needed: - khepera 3 robot  with korebot 2
 *          - khepera 3 gripper optionnal
 *          - libkorebot version >= 1.16                                       
 * 
 
 versions:
 1.0 December 2010: first release
 
 */

/*
  lua commands:
-- distance_inMeters=simExtK3_getInfrared(index_of_ir_sensor_0_to_8)
-- distance_inMeters=simExtK3_getUltrasonic(index_of_us_sensor_0_to_5)
-- intensity_0_to_1=simExtK3_getLineSensor(index_of_line_sensor_0_to_1)
-- distance_inMeters=simExtK3_getGripperProxSensor(index_of_finger_prox_sensor_0_to_1)

-- simExtK3_setVelocity(velocityLeft_radPerSec,velocityRight_radPerSec)
-- simExtK3_setArmPosition(position_300_to_900)
-- simExtK3_setGripperGap(gap_0_to_170)

-- simWait() [s]
-- simGetSimulationTime(-) [s]
-- simStopSimulation(-) : stopper 

\todo:

3 functions hors simulation (pour wireless)
-- ApplyNewScript
-- StopExecution
-- StartSimulation (restart too)
-- setparameters : pour ajuster entre hardware et software

lua_sethook()
lua_yield

*/


//#include "stdafx.h" // for windows
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#define ARM 1


extern "C" {
	#include "lua/lua.h"
	#include "lua/lauxlib.h"
	#include "lua/lualib.h"
  
  #ifdef ARM
 
  #include <korebot/korebot.h>
  #include "khepera3.h"
  #include "kgripper.h" 
  
  #endif
  
}

static int quitReq = 0;


/*--------------------------------------------------------------------*/
/*! Quit the program.
*/
int quit()
{
  quitReq = 1;
 // printf("quit requested...\n");
}



/*--------------------------------------------------------------------*/
/*! Reference time in milliseconds (initialized at the 
 *  application startup) 
 */
static unsigned long ref_time = 0;

/*--------------------------------------------------------------------*/
/*! 
 * This function gives a timestamp with a resolution of one 
 * millisecond that warp to zero 0 in about 49 days.
 *
 * \return the number of milliseconds since the start of the application 
 */
unsigned long getTimeMs(void)
{
  unsigned long val;
  struct timeval tv;

  gettimeofday( &tv , NULL );

  /* set the reference time */
  if ( ref_time == 0 ) {
    ref_time = (( tv.tv_sec * 1000 ) + ( tv.tv_usec / 1000 ));
  }

  val  = ( tv.tv_sec * 1000 ) + ( tv.tv_usec / 1000 );
  val -= ref_time;
 

  return val;
}

/*--------------------------------------------------------------------*/

// Lua functions


#ifdef ARM

// get infrared sensor value
// lua input: index of IR sensor 0 to 8
// lua ouput: distance in [m]
int _simExtK3_getInfrared(lua_State* L)
{ 
	int index;

	int numberOfArguments=lua_gettop(L);
	
	
	
	if (!lua_isnumber(L,1))
	{
		lua_pushnumber(L,-1); 
		return 1;
	}	
	
	index=lua_tointeger(L,1);
	
	if ((index < 0) || (index > 8)  )
	{
		lua_pushnumber(L,-1); 
		return 1;
	}
	 
	
	lua_pushnumber(L,log((double)proxIR(index)/4464.0)/-0.2613/100.0);

	return(1); // We return 1 values
}




// get infrared sensor value
// lua input: index of IR line sensor 0 to 1
// lua ouput: value from 0 to 1
int _simExtK3_getLineSensor(lua_State* L)
{ 
	int index;

	int numberOfArguments=lua_gettop(L);
	
	
	
	if (lua_isnumber(L,1))
	{
		index=lua_tointeger(L,1);
		
		if ((index < 0) || (index > 1)  )
		{
			lua_pushnumber(L,-1); 
			return 1;
		}
		
		
	}	
	else
	{
		lua_pushnumber(L,-1); 
		return 1;
	}	
	
	lua_pushnumber(L,(double)proxIR(index+9)/4096.0);  // input value, output value: value 0 to 1

	return(1); // We return 1 values
}

// get us sensor value
// lua input: index of US sensor 0 to 4
// lua ouput: distance in [m]
int _simExtK3_getUltrasonic(lua_State* L)
{ 
	int index;

	int numberOfArguments=lua_gettop(L);
	
	
	
	if (lua_isnumber(L,1))
	{
		index=lua_tointeger(L,1);
	}	
	else
	{
		lua_pushnumber(L,-1); 
		return 1;
	}
	
	
	if ((index < 0) || (index > 4)  )
	{
		lua_pushnumber(L,-1); 
		return 1;
	}	
	
	lua_pushnumber(L,measureUS(index)/100.0); // cm in m

	return(1); // We return 1 values
}

// get gripper gap sensor
// lua input: index of gripper proximity sensor 0 to 1
// lua ouput: distance in [m]
int _simExtK3_getGripperProxSensor(lua_State* L)
{
	int index;

	int numberOfArguments=lua_gettop(L);
	
	
	
	if (lua_isnumber(L,1))
	{
		index=lua_tointeger(L,1);
	}	
	else
	{
		lua_pushnumber(L,-1); 
		return 1;
	}	
	
	if ((index < 0) || (index > 1)  )
	{
		lua_pushnumber(L,-1); 
		return 1;
	}
	
	lua_pushnumber(L,get_gripper_sensor(index));

	return(1); // We return 1 values
}

// set velocity
// input : left wheel, righ wheel in [rad/s]
int _simExtK3_setVelocity(lua_State* L)
{ 
	double left_speed,right_speed;

	int numberOfArguments=lua_gettop(L);
	
	
	
	if (lua_isnumber(L,1))
	{
		left_speed=(double)lua_tonumber(L,1);
	} else 
	{
		return(0);	
	}
	
	if (lua_isnumber(L,2))
	{
		right_speed=(double)lua_tonumber(L,2);
	}
	else
	{
		return(0);
	}	
	

	// convert speed
	left_speed= left_speed*128.8*144.01;  // rad/s * radius * Kspeed
	right_speed= right_speed*128.8*144.01;  // rad/s * radius * Kspeed
	
	motSpeed((int)left_speed,(int)right_speed);
	
	
	return(0); // We return 0 values
}

// move arm position:
// input: 300 (down) to 900 (up)
int _simExtK3_setArmPosition(lua_State* L)
{
	int position;

	int numberOfArguments=lua_gettop(L);
	
	
	
	if (lua_isnumber(L,1))
	{
		position=(int)lua_tonumber(L,1);
	} else 
	{
		return(0);	
	}
	
	movearm(position);
	
	return(0);
}	


// open gripper
// input: 0 (closed) to 170 (open)
int _simExtK3_setGripperGap(lua_State* L)
{
	int position;


  int numberOfArguments=lua_gettop(L);
	
	if (lua_isnumber(L,1))
	{
		position=(int)lua_tonumber(L,1);
		
		
	} else 
	{
		return(0);	
	}
	
	
	printf("setGripperGap %d\n",position);
	movegrip(position);
	

	
	return(0);
}


#endif

// wait seconds [s] 
int _simWait(lua_State* L)
{ 
	int numberOfArguments=lua_gettop(L);
	
	if (lua_isnumber(L,1))
	{
		printf("wait %5.3f s\n",lua_tonumber(L,1));
		usleep((long int)(1000000.0*lua_tonumber(L,1)));				
		printf("waited\n");
	}
	
	
	return(0); // We return 0 values
}




int _simStopSimulation(lua_State* L)
{
	quit();
	return(0); // We return 0 values
}


// return the time in [s] from the start of the simulation
int _simGetSimulationTime(lua_State* L)
{
	lua_pushnumber(L,(double)(getTimeMs())/1000);

	return(1); // We return 1 values
}





// Following is a very simple example script that will be run:
#define SCRIPT_TEXT "for i=1,3,1 do simPrintHelloWorld('blabla',1324) end\n\
a,b=simPrintHelloWorld(1)\n\
print('The values returned are: '..a..' and '..b)"


// Following is the custom Lua function (which makes the link Lua - C/C++ program)
int _simPrintHelloWorld(lua_State* L)
{ // This is called from Lua
	// Here we can check the input arguments:
	int numberOfArguments=lua_gettop(L);
	for (int i=0;i<numberOfArguments;i++)
	{ // Just as an example:
		if (lua_isnumber(L,i+1))
			printf("Input argument #%i is a number\n",i+1);
		if (lua_isstring(L,i+1))
			printf("Input argument #%i is a string\n",i+1);
	}
	// Now we do what the function actually does: print "Hello World!":
	printf("Hello World!\n");

	// Now we return 2 values to Lua:
	lua_pushnumber(L,59); // First value (a number)
	lua_pushstring(L,"k-team"); // Second value (a string)
	return(2); // We return 2 values
}


/*---------------------------------------------------------------------------*/

// load the lua script
int load_script(char *filename, char *script[])
{
	FILE *pfile=NULL;
	
	
	long lSize;
  size_t result;
  
	/* open script file */
  if((pfile=fopen(filename,"r")) == NULL )
  {
    printf("Cannot open LUA script file ""%s"" !\n",filename);      
    return -1;
  }


	// obtain file size:
  fseek (pfile , 0 , SEEK_END);
  lSize = ftell (pfile);
  rewind (pfile);

	if (*script != NULL)
	{
		printf("script variable freeing\n");
		free(*script);
		*script=NULL;
		
	}

	printf("file size: %d bytes\n",lSize);

  // allocate memory to contain the whole file:
  *script = (char*) malloc (sizeof(char)*(lSize)+1);
  if (*script == NULL) 
  {
  	fputs ("Memory error\n",stderr);
  	return -2;
  }

  // copy the file into the buffer:
  result = fread (*script,1,lSize,pfile);
  if (result != lSize) 
  {
  	fputs ("Reading error\n",stderr);
  	return -3;
  }
  

	(*script)[lSize]='\0';

	
	fclose(pfile);

	
	return 0;
}





// lua function structure "type"
struct lua_fonctions_st {
	char *name;
	int (*function)(lua_State*);
};

#define LUA_FUNCTIONS_NB 11 // number of lua functions implemented here

// lua functions array
struct lua_fonctions_st lua_fonctions [] =
{
{"simPrintHelloWorld",_simPrintHelloWorld},
{"simWait",_simWait},
{"simGetSimulationTime",_simGetSimulationTime},
{"simStopSimulation",_simStopSimulation},
{"simExtK3_getInfrared",_simExtK3_getInfrared},
{"simExtK3_getUltrasonic",_simExtK3_getUltrasonic},
{"simExtK3_getLineSensor",_simExtK3_getLineSensor},
{"simExtK3_getGripperProxSensor",_simExtK3_getGripperProxSensor},
{"simExtK3_setVelocity",_simExtK3_setVelocity},
{"simExtK3_setArmPosition",_simExtK3_setArmPosition},
{"simExtK3_setGripperGap",_simExtK3_setGripperGap}
}; 






// ======================================================================

// main
int main(int argc, char* argv[])
{

	
	char *script=NULL;
	
	time_t file_time, file_time_old=0;
	
	lua_State* L;
	
	int luaerr;
	
	struct stat fst;
	
	bzero(&fst,sizeof(fst));
 
	
	#ifdef ARM
  // Set the libkorebot debug level - Highly recommended for development.
  kb_set_debug_level(2);

  //Init the korebot library
  if((kb_init( argc , argv )) < 0 )
  {
    printf("Error initialiasing libkorebot !\n");      
    return -1;
  }

	
	if(initKH3()<0)
	{
		printf("Error initialiasing robot !\n");      
    return -2;
	} 



	if(initGripper()<0)
	{
		printf("Error initialiasing gripper !\n");      
	}

	#endif
	
  // test parameter argument number
  if (argc != 2)
	{
		printf("No lua script in parameter!\n\n  synthax: ./kh3luaparser YOUR_LUA_SCRIPT\n\n  CTRL-C to quit\n");      
    return -4;
	}	 
	
	

	// execute until quit request 
	while (!quitReq)
	{
	
		// get script file modified time
		stat(argv[1],&fst);
		file_time=fst.st_mtime; 
		 
	
		// if different from previous time, reload the script file
		if (file_time_old != file_time)
		{
		
			printf("\nNew file time (or first time)!\n");
			/* open script file */
			if( load_script(argv[1],&script) != 0)
			{
				printf("Cannot open LUA script file ""%s"" !\n",argv[1]);      
				return -3;
			}
			
			printf("lua file loaded ok!\n");
		
			
			printf("*************** %s **********\n",argv[1]);
			printf("%s",script);
			printf("********************************************\n");
			
			// not first time
			if (file_time_old!=0)
			{
				// Cleanup:
			  lua_close(L);
				printf("lua closed (not first time)!\n");
		
			}
		
			file_time_old=file_time;
		
		 
	

	
		// Initialize a Lua state:
		L=lua_open();
		luaL_openlibs(L);
		
		
		// Register custom functions:
		
		for (int i=0;i<LUA_FUNCTIONS_NB; i++)
		{
			lua_register(L,lua_fonctions[i].name,lua_fonctions[i].function);
		}
		// lua_sethook(lua_hook)
			
		printf("functions registered!\n");	
			
		// Run a Lua script:
		// First load the text buffer (has to be done just once):
	  	if ((luaerr=luaL_loadbuffer(L,script,strlen(script),argv[1]))==0)
		//if (luaL_loadbuffer(L,SCRIPT_TEXT,strlen(SCRIPT_TEXT),argv[1])==0)
		{
		
				printf("script compiled ok !\n");
			 printf("----- start of the script output -----\n");
				// Now execute the program (can be called several times in a loop):
				//start_time=clock();
				ref_time=0;
				getTimeMs();
				if (lua_pcall(L,0,LUA_MULTRET,0)!=0)
				{
					printf("script execution: RUNTIME ERROR\n");
				} else
				{
				
					//clock_t endtim=clock();
					
					//double tim=((double)(endtim-start_time))/CLOCKS_PER_SEC ; 
					unsigned long tim=getTimeMs();
					printf("----- end of the script output -----\n");
					printf("script execution: SUCCESSFUL in %10ld [ms]\n\n",tim);
				}
	
		}
		else 
		{
			printf("COMPILATION ERROR: ");
			
			if (luaerr==LUA_ERRSYNTAX)
				printf("synthax error!\n");
			else if (luaerr==LUA_ERRMEM) 
				printf("memory allocation error!\n");	
			else
				printf("unknown error!\n");	
				
		}
		}
		
		printf("Waiting for modification of file \"%s\" ...\n",argv[1]);
		sleep(1); // wait 1 s before checking again
	} 

	// Cleanup:
	lua_close(L);
	
	
	free(script);	

	return 0;
}

